@@ -1,8 +1,26 @@
 #include <Python.h>
 #include <llvm/ADT/SmallVector.h>
-#include <llvm/Value.h>
-#include <llvm/DerivedTypes.h>
-#include <llvm/Function.h>
+#if LLVM_VERSION_MAJOR >= 3 && LLVM_VERSION_MINOR >= 3
+    #include <llvm/IR/Value.h>
+    #include <llvm/IR/DerivedTypes.h>
+    #include <llvm/IR/Function.h>
+    #include <llvm/IR/Module.h>
+    #include <llvm/IR/Constants.h>
+    #include <llvm/IR/Intrinsics.h>
+    #include <llvm/IR/IRBuilder.h>
+    #if LLVM_VERSION_MINOR >= 4
+        #include <llvm/Support/MemoryObject.h>
+        #include <llvm/MC/MCDisassembler.h>
+    #endif
+#else
+    #include <llvm/Value.h>
+    #include <llvm/DerivedTypes.h>
+    #include <llvm/Function.h>
+    #include <llvm/Module.h>
+    #include <llvm/Constants.h>
+    #include <llvm/Intrinsics.h>
+    #include <llvm/IRBuilder.h>
+#endif
 #include <llvm/Support/raw_ostream.h>
 #include <llvm/Support/FormattedStream.h>
 #include <llvm/Support/MemoryBuffer.h>
@@ -12,14 +30,11 @@
 #include <llvm/ExecutionEngine/ExecutionEngine.h>
 #include <llvm/ExecutionEngine/GenericValue.h>
 #include <llvm/Linker.h>
-#include <llvm/Module.h>
 #include <llvm/Analysis/Verifier.h>
-#include <llvm/Constants.h>
-#include <llvm/Intrinsics.h>
-#include <llvm/IRBuilder.h>
 #include <llvm/PassRegistry.h>
 #include <llvm/Support/Host.h>
 
+#include <llvm/ExecutionEngine/MCJIT.h> // to make MCJIT working
 
 #include "auto_pyobject.h"
 
@@ -56,6 +71,15 @@ namespace extra{
 
 }
 
+static
+PyObject* callwrite(PyObject* self, PyObject* arg)
+{
+    char meth[] = "write";
+    char fmt[] = "O";
+    return PyObject_CallMethod(self, meth, fmt, arg);
+}
+
+
 static
 PyObject* make_raw_ostream_for_printing(PyObject* self, PyObject* args)
 {
@@ -270,7 +294,12 @@ llvm::ExecutionEngine* ExecutionEngine_create(
     ExecutionEngine *ee = ExecutionEngine::create(M, ForceInterpreter,
                                                   &ErrorStr, OptLevel,
                                                   GVsWithCode);
-    PyFile_WriteString(ErrorStr.c_str(), errout);
+    auto_pyobject buf = PyBytes_FromString(ErrorStr.c_str());
+    if (errout && NULL == callwrite(errout, *buf)){
+        return NULL;
+    }
+
+//    PyFile_WriteString(ErrorStr.c_str(), errout);
     return ee;
 }
 
@@ -292,7 +321,11 @@ llvm::ExecutionEngine* ExecutionEngine_createJIT(
     std::string ErrorStr;
     ExecutionEngine *ee = ExecutionEngine::createJIT(M, &ErrorStr, JMM, OL,
                                                      GCsWithCode, RM, CMM);
-    PyFile_WriteString(ErrorStr.c_str(), errout);
+    auto_pyobject buf = PyBytes_FromString(ErrorStr.c_str());
+    if (errout && NULL == callwrite(errout, *buf)){
+        return NULL;
+    }
+//    PyFile_WriteString(ErrorStr.c_str(), errout);
     return ee;
 }
 
@@ -474,7 +507,7 @@ PyObject* llvm_ParseBitCodeFile(llvm::StringRef Buf, llvm::LLVMContext& Ctx,
         std::string ErrStr;
         M = ParseBitcodeFile(MB, Ctx, &ErrStr);
         auto_pyobject buf = PyBytes_FromString(ErrStr.c_str());
-        if (NULL == PyObject_CallMethod(FObj, "write", "O", *buf)){
+        if (NULL == callwrite(FObj, *buf)){
             return NULL;
         }
 //        if (-1 == PyFile_WriteString(ErrStr.c_str(), FObj)) {
@@ -498,7 +531,7 @@ PyObject* llvm_WriteBitcodeToFile(const llvm::Module *M, PyObject* FObj)
     rso.flush();
     StringRef ref = rso.str();
     auto_pyobject buf = PyBytes_FromStringAndSize(ref.data(), ref.size());
-    return PyObject_CallMethod(FObj, "write", "O", *buf);
+    return callwrite(FObj, *buf);
 }
 
 static
@@ -512,9 +545,13 @@ PyObject* llvm_getBitcodeTargetTriple(llvm::StringRef Buf,
     if (FObj) {
         std::string ErrStr;
         Triple = getBitcodeTargetTriple(MB, Ctx, &ErrStr);
-        if (-1 == PyFile_WriteString(ErrStr.c_str(), FObj)) {
+        auto_pyobject buf = PyBytes_FromString(ErrStr.c_str());
+        if (NULL == callwrite(FObj, *buf)){
             return NULL;
         }
+//        if (-1 == PyFile_WriteString(ErrStr.c_str(), FObj)) {
+//            return NULL;
+//        }
     } else {
         Triple = getBitcodeTargetTriple(MB, Ctx);
     }
@@ -581,11 +618,17 @@ PyObject* Linker_LinkInModule(llvm::Linker* Linker,
                                PyObject* ErrMsg)
 {
     std::string errmsg;
+#if LLVM_VERSION_MAJOR >= 3 && LLVM_VERSION_MINOR >= 3
+    bool failed = Linker->linkInModule(Mod, &errmsg);
+#else
     bool failed = Linker->LinkInModule(Mod, &errmsg);
+#endif
     if (! failed) {
         Py_RETURN_FALSE;
     } else {
-        if (-1 == PyFile_WriteString(errmsg.c_str(), ErrMsg)) {
+    
+        auto_pyobject buf = PyBytes_FromString(errmsg.c_str());
+        if (NULL == callwrite(ErrMsg, *buf)){
             return NULL;
         }
         Py_RETURN_TRUE;
@@ -603,9 +646,13 @@ PyObject* Linker_LinkModules(llvm::Module* Dest,
     if (! failed) {
         Py_RETURN_FALSE;
     } else {
-        if (-1 == PyFile_WriteString(errmsg.c_str(), ErrMsg)) {
+        auto_pyobject buf = PyBytes_FromString(errmsg.c_str());
+        if (NULL == callwrite(ErrMsg, *buf)){
             return NULL;
         }
+//        if (-1 == PyFile_WriteString(errmsg.c_str(), ErrMsg)) {
+//            return NULL;
+//        }
         Py_RETURN_TRUE;
     }
 }
@@ -665,9 +712,14 @@ PyObject* llvm_verifyModule(const llvm::Module& Fn,
     bool failed = llvm::verifyModule(Fn, Action, &errmsg);
 
     if (failed) {
-        if (-1 == PyFile_WriteString(errmsg.c_str(), ErrMsg)) {
+        auto_pyobject buf = PyBytes_FromString(errmsg.c_str());
+        if (NULL == callwrite(ErrMsg, *buf)){
             return NULL;
         }
+
+//        if (-1 == PyFile_WriteString(errmsg.c_str(), ErrMsg)) {
+//            return NULL;
+//        }
         Py_RETURN_TRUE;
     } else {
         Py_RETURN_FALSE;
@@ -790,9 +842,13 @@ PyObject* DynamicLibrary_LoadLibraryPermanently(const char * Filename,
         std::string errmsg;
         failed = DynamicLibrary::LoadLibraryPermanently(Filename, &errmsg);
         if (failed) {
-            if (-1 == PyFile_WriteString(errmsg.c_str(), ErrMsg)) {
+            auto_pyobject buf = PyBytes_FromString(errmsg.c_str());
+            if (NULL == callwrite(ErrMsg, *buf)){
                 return NULL;
             }
+//            if (-1 == PyFile_WriteString(errmsg.c_str(), ErrMsg)) {
+//                return NULL;
+//            }
         }
     } else {
         failed = DynamicLibrary::LoadLibraryPermanently(Filename);
@@ -835,7 +891,12 @@ PyObject* TargetRegistry_lookupTarget(const std::string &Triple,
     std::string error;
     const Target* target = TargetRegistry::lookupTarget(Triple, error);
     if (!target) {
-        PyFile_WriteString(error.c_str(), Error);
+//        PyFile_WriteString(error.c_str(), Error);
+        auto_pyobject buf = PyBytes_FromString(error.c_str());
+        if (NULL == callwrite(Error, *buf)){
+            return NULL;
+        }
+
         Py_RETURN_NONE;
     } else {
         return pycapsule_new(const_cast<Target*>(target), "llvm::Target");
@@ -852,14 +913,18 @@ PyObject* TargetRegistry_lookupTarget(const std::string &Arch,
     std::string error;
     const Target* target = TargetRegistry::lookupTarget(Arch, Triple, error);
     if (!target) {
-        PyFile_WriteString(error.c_str(), Error);
+//        PyFile_WriteString(error.c_str(), Error);
+        auto_pyobject buf = PyBytes_FromString(error.c_str());
+        if (NULL == callwrite(Error, *buf)){
+            return NULL;
+        }
+
         Py_RETURN_NONE;
     } else {
         return pycapsule_new(const_cast<Target*>(target), "llvm::Target");
     }
 }
 
-
 static
 PyObject* TargetRegistry_getClosestTargetForJIT(PyObject* Error)
 {
@@ -867,7 +932,12 @@ PyObject* TargetRegistry_getClosestTargetForJIT(PyObject* Error)
     std::string error;
     const Target* target = TargetRegistry::getClosestTargetForJIT(error);
     if (!target) {
-        PyFile_WriteString(error.c_str(), Error);
+        auto_pyobject buf = PyBytes_FromString(error.c_str());
+        if (NULL == callwrite(Error, *buf)){
+            return NULL;
+        }
+
+//        PyFile_WriteString(error.c_str(), Error);
         Py_RETURN_NONE;
     } else {
         return pycapsule_new(const_cast<Target*>(target), "llvm::Target");
@@ -875,6 +945,66 @@ PyObject* TargetRegistry_getClosestTargetForJIT(PyObject* Error)
 
 }
 
+static
+PyObject* TargetRegistry_targets_list()
+{
+    using namespace llvm;
+
+    return iterator_to_pylist_deref<TargetRegistry::iterator>(
+                  TargetRegistry::begin(), TargetRegistry::end(),
+                  "llvm::Target", "llvm::Target");
+}
+
+#if LLVM_VERSION_MAJOR >= 3 && LLVM_VERSION_MINOR >= 4
+static
+PyObject* MemoryObject_readBytes(const llvm::MemoryObject *mobj,
+                                 uint64_t addr,
+                                 uint64_t size
+                                 )
+{
+    int status;
+    uint8_t *bytes;
+    PyObject* po;
+
+    if(size < 1)
+        goto fail;
+
+    bytes = new uint8_t[size];
+    if(bytes == NULL)
+        goto fail;
+
+    status = mobj->readBytes(addr, size, bytes);
+    if(status != 0) {
+        delete bytes;
+        goto fail;
+    }
+
+    po = PyBytes_FromStringAndSize((const char *) bytes, size);
+    delete bytes;
+
+    return po;
+fail:
+    Py_RETURN_NONE;
+}
+
+static
+PyObject* MCDisassembler_getInstruction(llvm::MCDisassembler *disasm, 
+                                        llvm::MCInst &instr,
+                                        const llvm::MemoryObject &region,
+                                        uint64_t address
+                                        )
+{
+    uint64_t size;
+    llvm::MCDisassembler::DecodeStatus status;
+
+    size = 0;
+    status = disasm->getInstruction(instr, size, region, address, 
+                                    llvm::nulls(), llvm::nulls());
+    return Py_BuildValue("(i,i)", int(status), size);
+}
+
+#endif /* llvm >= 3.4 */
+
 static
 PyObject* llvm_sys_getHostCPUFeatures(PyObject* Features)
 {
@@ -898,3 +1028,23 @@ PyObject* llvm_sys_getHostCPUFeatures(PyObject* Features)
     }
 }
 
+#if LLVM_VERSION_MAJOR >= 3 && LLVM_VERSION_MINOR >= 3
+static
+PyObject* llvm_sys_isLittleEndianHost()
+{
+    if (llvm::sys::IsLittleEndianHost)
+        Py_RETURN_TRUE;
+    else
+        Py_RETURN_FALSE;
+}
+
+static
+PyObject* llvm_sys_isBigEndianHost()
+{
+    if (llvm::sys::IsBigEndianHost)
+        Py_RETURN_TRUE;
+    else
+        Py_RETURN_FALSE;
+}
+#endif 
+
